import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import *
from pytorch_mixtures.utils import Experts


class MoELayer(nn.Module):
    def __init__(self, num_experts: int, router: nn.Module, experts: List[nn.Module], capacity_factor: float) -> None:
        super().__init__()
        self.router = router
        self.num_experts = num_experts
        self.capacity_factor = capacity_factor
        self.experts = Experts(num_experts, experts)
    
    def forward(self, token_inputs):
        [B, N, D] = token_inputs.shape
        expert_capacity = int((N * self.capacity_factor) // self.num_experts)
        # send tokens to router
        routing_instructions = self.router(token_inputs, expert_capacity)
        # dispatch to experts
        expert_inputs = torch.einsum("bnd,bnec->becd", token_inputs, routing_instructions.dispatch_tensor)
        # processing by experts
        expert_outputs = self.experts(expert_inputs)
        # combine expert outputs
        output = torch.einsum("becd,bnec->bnd", expert_outputs, routing_instructions.combine_tensor)
        return output
 


# Reference:
# 1. Google Flaxformer GitHub: https://github.com/google/flaxformer

import torch
import torch.nn as nn
from torch import Tensor
import torch.nn.functional as F
from typing import *
from pytorch_mixtures.utils import load_balancing_loss, router_z_loss, _one_hot


class RouterOutput():
    def __init__(self, dispatch_tensor: Tensor, combine_tensor: Tensor, aux_loss: Tensor, router_z_loss: Tensor) -> None:
        self.dispatch_tensor = dispatch_tensor
        self.combine_tensor = combine_tensor
        self.aux_loss = aux_loss
        self.router_z_loss = router_z_loss


class RouterWeights(nn.Module):
    def __init__(self, dim: int, num_experts: int) -> None:
        super().__init__()
        self.w_gate = nn.Parameter(torch.randn(dim, num_experts))

    def forward(self, x: Tensor) -> Tensor:
        return torch.einsum("bnd,de->bne", x, self.w_gate)


class Router(nn.Module):
    def __init__(self, dim: int, num_experts: int) -> None:
        super().__init__()
        self.weights = RouterWeights(dim, num_experts)

    def forward(self, token_inputs: Tensor, expert_capacity: int) -> RouterOutput:
        router_logits = self.weights(token_inputs)
        z_loss = router_z_loss(router_logits)
        router_probs = F.softmax(router_logits, dim=-1)

        routing_instructions = self.compute_routing_instructions(router_probs, expert_capacity)
        routing_instructions.update({"router_z_loss": z_loss})

        return RouterOutput(**routing_instructions)


class ExpertChoiceRouter(Router):
    def compute_routing_instructions(self, router_probs: Tensor, expert_capacity: int) -> dict:
        [B, N, E] = router_probs.shape
        # shape = [B, E, N]
        transposed_router_probs = torch.permute(router_probs, (0, 2, 1))
        # shape = [B, E, C]
        expert_gate, expert_index = torch.topk(transposed_router_probs, k=expert_capacity, dim=-1)
        # make the dispatch tensor
        # shape = [B, E, C, N]
        dispatch_tensor = _one_hot(expert_index, N)
        # shape = [B, N, E, C]
        dispatch_tensor = torch.permute(dispatch_tensor, (0, 3, 1, 2)).to(torch.float32)

        # make the combine tensor
        # shape = [B, N, E, C]
        combine_tensor = torch.einsum(
			"bec,bnec->bnec",
			expert_gate,
			dispatch_tensor
		).to(torch.float32)
        
        aux_loss = torch.tensor(0.0).to(router_probs.device)
        return {"dispatch_tensor": dispatch_tensor, "combine_tensor": combine_tensor, "aux_loss": aux_loss}


class TopkRouter(Router):
    def __init__(self, dim: int, num_experts: int, topk: int) -> None:
        # self.dim = dim
        # self.num_experts = num_experts
        super().__init__(dim, num_experts)
        self.topk = topk

    def compute_routing_instructions(self, router_probs: Tensor, expert_capacity: int) -> dict:
        [B, N, E] = router_probs.shape
        expert_gate, expert_indices = router_probs.topk(k=self.topk, dim=-1)

        aux_loss = load_balancing_loss(router_probs, expert_indices)

        # shape: [B, self.topk, N]
        expert_indices = torch.permute(expert_indices, (0, 2, 1))
        # shape: [B, self.topk * N]
        expert_indices = expert_indices.reshape(B, self.topk * N)

        # shape: [B, N * self.topk, E]
        expert_mask = _one_hot(expert_indices, E).to(torch.int32)
        # shape: [B, self.topk * N, E]
        token_priority = torch.cumsum(expert_mask, dim=1) * expert_mask - 1.0
        # shape: [B, self.topk, N, E]
        token_priority = token_priority.view(B, self.topk, N, E)
        # shape: [B, N, self.topk, E]
        token_priority = torch.permute(token_priority, (0, 2, 1, 3))
        # shape: [B, N, E]
        token_priority, _ = torch.max(token_priority, dim=2)
        token_priority = token_priority.long()

        # make the dispatch and combine tensors
        # shape: [B, N, E, expert_capacity]
        dispatch_tensor = _one_hot(token_priority, expert_capacity).float()
        # shape: [B, N, E, expert_capacity]
        combine_tensor = torch.einsum(
            "...te,...tec->...tec",
            router_probs,
            dispatch_tensor
        )
        return {"dispatch_tensor": dispatch_tensor, "combine_tensor": combine_tensor, "aux_loss": aux_loss}



from .moe_layer import MoELayer
from .mod_layer import MoDLayer
from .routing import ExpertChoiceRouter, TopkRouter
from .utils import MHSA

import torch
import torch.nn as nn

from absl.testing import absltest
from absl.testing import parameterized


def difference_within_bound(a, b, eps=1e-4):
    x = a-b
    return x.all() < eps

def print_report(a, b):
    bounds = [1e-5, 1e-6, 1e-7]
    best_bound = None
    for ep in bounds:
        result = difference_within_bound(a, b, eps=ep)
        if result:
            best_bound = ep
        
    if best_bound is not None:
        print(f"The difference between the two tensors is 0 within {best_bound} across across elements.")
        return True
    else:
        print(f"The two tensors cannot be reliably called equal.")
        return False


class PyTorchMixturesTests(parameterized.TestCase):
    @torch.no_grad()
    def test_expert_choice_moe_with_identical_routers(self):
        B, N, D = 1, 128, 64
        E = 8
        C = 1
        expert_capacity = int((N * C) // E)
        x = torch.randn(B, N, D)

        # set up an expert to use as reference
        reference_expert = nn.Linear(64, 64)
        many_experts = [nn.Linear(64, 64) for _ in range(E)]
        # make all the experts the same as the reference expert
        for i in range(E):
            many_experts[i].load_state_dict(reference_expert.state_dict())
        
        router = ExpertChoiceRouter(dim=64, num_experts=E)
        moe = MoELayer(num_experts=E, router=router, experts=many_experts, capacity_factor=C)
        y = moe(x)
        z = reference_expert(x)

        print(" ")
        print("Test details:")
        print("--------------------------------------------------------------------------")
        print("1. All experts are initialized as the same network.")
        print("2. The router follows Expert-Choice protocol and is randomy initialized.")
        print("3. Since all experts are equal, the weighted sum of the expert outputs")
        print("   should equal the input tensor passed through any expert.")
        print("4. This test reports whether the different b/w the MoE output and the")
        print("   input tensor passed through any expert is 0 (sufficiently).")
        print(" ")
        
        report_return = print_report(y, z)
        self.assertEqual(report_return, True)

    @torch.no_grad()
    def test_topk_moe_with_identical_routers(self):
        B, N, D = 1, 128, 64
        E = 8
        C = 1
        K = 2
        expert_capacity = int((N * C) // E)
        x = torch.randn(B, N, D)

        # set up an expert to use as reference
        reference_expert = nn.Linear(64, 64)
        many_experts = [nn.Linear(64, 64) for _ in range(E)]
        # make all the experts the same as the reference expert
        for i in range(E):
            many_experts[i].load_state_dict(reference_expert.state_dict())
        
        router = TopkRouter(dim=64, num_experts=E, topk=K)
        moe = MoELayer(num_experts=E, router=router, experts=many_experts, capacity_factor=C)
        y = moe(x)
        z = reference_expert(x)

        print(" ")
        print("Test details:")
        print("--------------------------------------------------------------------------")
        print("1. All experts are initialized as the same network.")
        print("2. The router follows Top-2 protocol and is randomy initialized.")
        print("3. Since all experts are equal, the weighted sum of the expert outputs")
        print("   should equal the input tensor passed through any expert.")
        print("4. This test reports whether the different b/w the MoE output and the")
        print("   input tensor passed through any expert is 0 (sufficiently).")
        print(" ")
        
        report_return = print_report(y, z)
        self.assertEqual(report_return, True)

    @torch.no_grad()
    def test_expert_choice_mod(self):
        B, N, D = 1, 128, 64
        E = 8
        C = 1
        expert_capacity = int((N * C) // E)
        x = torch.randn(B, N, D)
        
        router = ExpertChoiceRouter(dim=D, num_experts=1)
        attn_fn = MHSA(dim=D, num_heads=E)
        mod = MoDLayer(router=router, attn_fn=attn_fn, capacity_factor=C)
        y = mod(x)

        print(" ")
        print("Test details:")
        print("-----------------------------------------------------------------")
        print("Only checks for an error-free forward pass through the MoD layer.")
        print(" ")

def run_tests():
    absltest.main(__name__)



# Like routing to one expert which is a attention/transformer block
# Expert Choice Routing
# NOTE: This has not been tested via training.

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from typing import *


class MoDLayer(nn.Module):
    def __init__(self, router: nn.Module, attn_fn: nn.Module, capacity_factor: float) -> None:
        super().__init__()
        self.router = router
        self.capacity_factor = capacity_factor
        self.attn_fn = attn_fn
    
    def forward(self, token_inputs: Tensor) -> Tensor:
        [B, N, D] = token_inputs.shape
        expert_capacity = int((N * self.capacity_factor))
        # send tokens to router
        routing_instructions = self.router(token_inputs, expert_capacity)
        # dispatch to experts
        attn_inputs = torch.einsum("bnd,bnec->becd", token_inputs, routing_instructions.dispatch_tensor)
        # processing by attention layer
        attn_inputs = attn_inputs.squeeze(1)
        attn_outputs = self.attn_fn(attn_inputs)
        attn_outputs = attn_outputs.unsqueeze(1)
        # combine expert outputs
        output = torch.einsum("becd,bnec->bnd", attn_outputs, routing_instructions.combine_tensor)
        return output



# References: 
# 1. Lucidrains ST-MoE Github: https://github.com/lucidrains/st-moe-pytorch
# 2. Google Flaxformer GitHub: https://github.com/google/flaxformer

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from typing import *
from einops import rearrange


class RMSNorm(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.scale = dim ** 0.5
        self.gamma = nn.Parameter(torch.ones(dim))
    
    def forward(self, x: Tensor) -> Tensor:
        return F.normalize(x, dim=-1) * self.gamma * self.scale
    

class GEGLU(nn.Module):
    def __init__(self, dim: int, mult_bias: bool = True) -> None:
        super().__init__()
        self.mult_bias = nn.Parameter(torch.ones(dim)) if mult_bias else 1
    
    def forward(self, x: Tensor) -> Tensor:
        x, gate = x.chunk(2, dim=-1)
        return F.gelu(gate) * x * self.mult_bias


class Experts(nn.Module):
    def __init__(self, num_experts: int, experts: List[nn.Module]) -> None:
        super().__init__()
        self.num_experts = num_experts
        self.experts = nn.ModuleList(experts)
    
    def forward(self, x: Tensor) -> Tensor:
        [B, E, N, D] = x.shape
        x = torch.permute(x, (1, 0, 2, 3))
        outputs = []

        for i in range(self.num_experts):
            per_expert_output = self.experts[i](x[i])
            outputs.append(per_expert_output.unsqueeze(0))

        outputs = torch.stack(outputs).view(E, B, N, D)
        return outputs.view(B, E, N, D)


class MHSA(nn.Module):
    def __init__(self, dim: int, num_heads: int, scaling: bool = True, dropout: float = 0.0) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.embedding_dim = dim
        self.dropout = nn.Dropout(dropout)
        self.scaling = scaling

        self.queries_keys_values = nn.Linear(dim, 3*dim)
        self.projection = nn.Linear(dim, dim)

    def forward(self, x: Tensor) -> Tensor:
        splits = rearrange(self.queries_keys_values(
            x), 'b n (h d qkv) -> (qkv) b h n d', qkv=3, h=self.num_heads)
        queries, keys, values = splits[0], splits[1], splits[2]

        attention = torch.einsum('bhqd, bhkd -> bhqk', queries, keys)
        attention = nn.functional.softmax(attention, dim=-1)
        if self.scaling:
            attention = attention / (self.embedding_dim**0.5)

        attention = self.dropout(attention)

        output = torch.einsum('bhad, bhdv -> bhav', attention, values)
        output = rearrange(output, 'b h a v -> b a (h v)')
        return self.projection(output)



def load_balancing_loss(router_probs: Tensor, expert_indices: Tensor) -> Tensor:
    [B, N, E] = router_probs.shape
    expert_mask = F.one_hot(expert_indices, E)
    expert_mask, _ = expert_mask.max(dim=-2)
    
    tokens_per_group_and_expert = expert_mask.float().mean(dim=-2).float()
    router_probs_per_group_and_expert = router_probs.mean(dim=-2).float()
    
    loss = (tokens_per_group_and_expert * router_probs_per_group_and_expert) * E**2
    return loss


def router_z_loss(router_logits: Tensor) -> Tensor:
    [B, N, E] = router_logits.shape
    log_z = torch.logsumexp(router_logits, dim=-1)
    loss = log_z**2
    return loss.sum().float() / (B * N) 


def _one_hot(indices: Tensor, num_classes: int, dtype=torch.int32) -> Tensor:
    """
    Augments the default `torch.nn.functional.one_hot` as follows:
    1. ignores negative elements in the input `indices`
    2. ignores elements with value greater than `num_classes-1`.
    """
    mask1 = indices < 0
    mask2 = indices >= num_classes
    new_indices = indices
    new_indices[mask1] = 0
    new_indices[mask2] = 0
    res = F.one_hot(new_indices, num_classes)
    res[mask1] = 0
    res[mask2] = 0
    return res.to(dtype)



import setuptools


def parse_requirements(filename):
    with open(filename, 'r') as f:
        return [line.strip() for line in f if line and not line.startswith('#')]

setuptools.setup(
    name="pytorch-mixtures",
    version="0.1.1",
    author="Jaisidh Singh",
    author_email="jaisidhsingh@gmail.com",
    description="The one-stop solution to easily integrate MoE & MoD layers into custom PyTorch code.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/jaisidhsingh/pytorch-mixtures",
    packages=setuptools.find_packages(),
    install_requires=parse_requirements('requirements.txt'),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)
